.header	msxdos

; CS8900A I/O ports
.def	PPPntrH		$0b
.def	PPPntrL		$0a
.def	PPDataH		$0d
.def	PPDataL		$0c
.def	BDOS		$0005

; load DE with pointer to string
.macro printf
	ld %c,$09	; PutString BDOS call
	call BDOS
.endm

.macro	setRegister	; points the PP pointer to BC
	ld %a,%b
	out $0b,%a	; high byte
	ld %a,%c
	out $0a,%a	; low byte
.endm

; this macro reads into DE the data pointed to by BC
.macro	readData
	ld %a,%b
	out $0b,%a	; high byte
	ld %a,%c
	out $0a,%a	; low byte

	in %a,$0d	; loads data
	ld %e,%a
	in %a,$0c
	ld %d,%a
.endm

; this macro writes DE at location pointed to by BC
.macro	writeData
	ld %a,%b
	out $0b,%a	; high byte
	ld %a,%c
	out $0a,%a	; low byte

	ld %a,%d
	out $0d,%a
	ld %a,%e
	out $0c,%a
.endm

detect:
	ld %bc, 0	; point packet page pointer to 0x0000
	
	.x readData

	; compare value
	ld %hl,$0E63
	sbc %hl,%de

	jr z, detect1

	ld %de, error1
	.x printf
	ret

detect1:
	ld %de, found
	.x printf

; Set the MAC address
	ld %bc,$0158	; Internal address register

	ld %de,$def0
	.x writeData
	inc %c
	inc %c
	
	ld %de,$00ba
	.x writeData
	inc %c
	inc %c
	
	ld %de,$0100
	.x writeData

	ld %de, mac
	.x printf

; allow packets
	ld %bc,$0104
	.x setRegister

	ld %a,#00001101
	out $0d,%a

; enable receiver and transmitter
	ld %bc,$0112
	.x setRegister

	ld %a,#11000000	; set SerTxOn and SerRxOn
	out $0c,%a

;	ld %de, arp		; sends a dummy response
;	ld %bc, 42
;	call sendframe

; this routine loops until it receives a packet
poll:
	ld %bc,$0124	; RxEvent register
	.x setRegister

_poll:
	in %a,$0d
	and 1		; check RxOK flag
	jr z, _poll

; lets the user know it received a frame
	ld %de, gotf
	.x printf

readframe:
; reads the frame
	in %a, 1	; reads RxStatus
;	ld %d, %a
	in %a, 0
;	ld %e, %a	; ignore values for now

	in %a, 1	; reads RxLength
	ld %b, %a
	in %a, 0
	ld %c, %a
	push %bc

	srl %b		; divide length by 2 to simplify loop
	rr %c

	ld %hl, recvbf
	ld %ix, recvbf

_readframe:
	in %a, 1	; copies a word
	ld (%hl), %a
	inc %hl

	in %a, 0
	ld (%hl), %a
	inc %hl

	dec %bc
;	dec %bc

	ld %a,%c
	or %b

	jr nz, _readframe
	pop %bc

decode:
; here it decodes the packet
	ld %a, (ix + 13)		; loads frame type
	cp $08
	jr nz, _decodee

	ld %a, (ix + 12)
	cp $00
	jr z, _decodei

	cp $06
	jr z, _decodea

	cp $35
	jr z, _decoder 

_decodee:
	ld %de, ferr
	.x printf
	jr poll

_decodei:
	ld %de, fip
	.x printf
	jr poll

_decodea:
	
	ld %de, farp
	.x printf

	ld %de, arp		; sends a dummy response
	ld %bc, 42
	call sendframe

	jr poll

_decoder:
	ld %de, farpr
	.x printf
	jp poll

	ret

; sendframe
;
; arguments:
;	DE	pointer to packet_data
;	BC	size of packet
;
; returns:
;	nothing

sendframe:
; bid for buffer space by writing command and frame length
	ld %h, %b
	ld %l, %c

	ld %c, $04
	ld %a, $c0	; write 00C0h command to TxCMD
	out (%c),%a
	inc %c
	ld %a, $00
	out (%c),%a

	inc %c		; write TxLength
	out (%c),%l
	inc %c
	out (%c),%h
	
; poll for space
	ld %bc,$0138	; BusST register
	.x setRegister

_sendframe:
	in %a,$0d
	and %a,$01
	jp z, _sendframe

; send frame
__sendframe:
	ld %a,(%de)
	out 0, %a
	inc %de

	ld %a,(%de)
	out 1, %a
	inc %de

	dec %hl
	dec %hl

	ld %a,%h
	or %l

	jr nz, __sendframe
	ret

found:	.ds	"Found CS8900A at I/O address 00h\n\r$"
mac:	.ds	"Setting MAC address to f0:de:ba:00:00:01\n\r$"
gotf:	.ds	"Received frame type: $"
fip:	.ds	"IP datagram\n\r$"
farp:	.ds	"ARP request\n\r$"
farpr:	.ds	"ARP response\n\r"
ferr:	.ds	"Invalid\n\r$"
error1:	.ds	"Error: no ethernet controller found\n\r$"

arp:
	.db $00 $07 $95 $bb $3e $8c
hwaddr:
	.db $f0 $de $ba $00 $00 $01
	.db $08 $06 $00 $01 $08 $00 $06 $04 $00 $02
_hwaddr:
	.db $f0 $de $ba $00 $00 $01
ipaddr:
	.db $c0 $a8 $01 $09
thwaddr:
	.db $00 $07 $95 $bb $3e $8c
tipaddr:
	.db $c0 $a8 $01 $02

; here is the receive buffer
; FIXME: mallocar isso aqui.
recvbf: .empty	$0600
