#ifndef UZIX_MODULE
#include <stdio.h>
#include <malloc.h>
#include "cs8900.h"
#endif

struct ethernet_frame tx_buffer;
struct ethernet_frame rx_buffer;

/* have this more or less hard-coded in advance */
uint8_t arp_request[42] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xf0, 0xde, 0xba, 0x00, 0x00, 0x01,
	0x08, 0x06, 0x00, 0x01, 0x08, 0x00,
	0x06, 0x04, 0x00, 0x01, 0xf0, 0xde,
	0xba, 0x00, 0x00, 0x01, 0xc0, 0xa8,
	0x01, 0x09, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0x00, 0x00, 0x00, 0x00 };

uint8_t arp_reply[42] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xf0, 0xde, 0xba, 0x00, 0x00, 0x01,
	0x08, 0x06, 0x00, 0x01, 0x08, 0x00,
	0x06, 0x04, 0x00, 0x02, 0xf0, 0xde,
	0xba, 0x00, 0x00, 0x01, 0xc0, 0xa8,
	0x01, 0x09, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

uint8_t ethernet_header[14] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xf0, 0xde, 0xba, 0x00, 0x00, 0x01,
	0x08, 0x00};

struct arp_packet *request = (struct arp_packet *) (arp_request + 14);
struct arp_packet *reply = (struct arp_packet *) (arp_reply + 14);

/* to save time, we store the IP address on the pre-prepared ARP packet above */
uint8_t *ip_addr = arp_reply + 28;
uint8_t *hw_addr = arp_reply + 6;
uint8_t netmask[4];
uint8_t subnet[4];
uint8_t gateway[4];

#define MAX_REC_BUF	8		/* must be power of 2!!! */

uint8_t *rec_buff[MAX_REC_BUF];
uint16_t rec_buff_sizes[MAX_REC_BUF];
uint8_t rec_buff_idx = 0;
uint8_t rec_buff_last = 0;

/*  Searches through table to see if ARP entry exists.  If it does then
 * MAC address is updated.  If a entry does not exist, then add in an
 * appropiate spot.
 * Depends on :
 *  1.) static time
 *  2.) #define MAX_ENTRIES
 * Be sure "time" is initialized to 0
 */
void add_arp_entry(uint8_t * hw_addr, uint8_t * ip_addr )
{
	/* Variables used for finding oldest entry */
	int longest_time=0;
	int temp_index; 
	/*Counters */
	int i=0;
	int j=0;

	if (((ip_addr[0] & netmask[0]) != subnet[0]) ||
			((ip_addr[1] & netmask[1]) != subnet[1]) ||
			((ip_addr[2] & netmask[2]) != subnet[2]))
		return;
	
	mytime++;

	for(i=0; i < ARP_CACHE_SIZE; i++){
		/* Check in reverse order to speed up search */
		for(j=3; j>=0; j--){
			if (ip_addr[j] != arp_cache[i].ip_addr[j]) break;
			
			/* If all 4 terms of the IP matched, then match will = 4. Update the MAC addr */
			memcpy(arp_cache[i].hw_addr, hw_addr, 6);
			arp_cache[i].timer = mytime;
			return;
		}
	}

	/*try to find a 0*/
	for(i=0; i < ARP_CACHE_SIZE; i++)
		if(	arp_cache[i].ip_addr[0] == 0 &&	arp_cache[i].ip_addr[1] == 0
			&& arp_cache[i].ip_addr[2] == 0 && arp_cache[i].ip_addr[3] == 0)
			goto replace;

	if(i == ARP_CACHE_SIZE)
	{
		/* if the program made it here, there are no empty slots.  Replace oldest.*/
		for(i=0; i < ARP_CACHE_SIZE; i++)
		{
			if(mytime - arp_cache[i].timer > longest_time)
			{
				longest_time = mytime - arp_cache[i].timer;
				temp_index = i;
			}
		}
		i = temp_index;
	}

replace:
	/* "i" now is the index to the oldest or blank entry. Replace it.*/
	memcpy(arp_cache[i].ip_addr, ip_addr, 4);
	memcpy(arp_cache[i].hw_addr, hw_addr, 6);
	arp_cache[i].timer = mytime;

	if (tx_buffer.valid = 1)
		if ( ((struct ip_packet *) tx_buffer.data)->target_ip_addr[3] == ip_addr[3] &&
			((struct ip_packet *) tx_buffer.data)->target_ip_addr[2] == ip_addr[2] &&
			((struct ip_packet *) tx_buffer.data)->target_ip_addr[1] == ip_addr[1] &&
			((struct ip_packet *) tx_buffer.data)->target_ip_addr[0] == ip_addr[0])
		{
			/* send packet that was waiting in buffer */
			memcpy(ethernet_header, ip_addr, 6);
			writeFrame(tx_buffer.data, tx_buffer.length);
			tx_buffer.valid = 0;
		}

	return;
}

void do_arp_reply (struct ethernet_frame *frame)
{
	struct arp_packet *arp_frame = (struct arp_packet *) (frame->data + 14);

	if (arp_frame->target_ip_addr[3] != ip_addr[3])
		return;
	else if (arp_frame->target_ip_addr[2] != ip_addr[2])
		return;
	else if (arp_frame->target_ip_addr[1] != ip_addr[1])
		return;
	else if (arp_frame->target_ip_addr[0] != ip_addr[0])
		return;

	/* FIXME: send_packet should fill in the target HW address */
	memcpy(arp_reply + 32, frame->data + 6, 6);
	memcpy(arp_reply + 0, frame->data + 6, 6);
	memcpy(reply->target_ip_addr, arp_frame->source_ip_addr, 4);

	sendFrame(arp_reply, 42);

	add_arp_entry(arp_frame->source_hw_addr, arp_frame->source_ip_addr);

	return;
}

void do_arp_request (uint8_t *ip)
{
	memcpy(request->target_ip_addr, ip, 4);
	sendFrame(arp_request, 42);

	return;
}

uint8_t * resolve_ip(uint8_t *ip)
{
	uint8_t i;
	
	for (i = 0 ; i < ARP_CACHE_SIZE; i++)
		if (	arp_cache[i].ip_addr[3] == ip[3] &&
			arp_cache[i].ip_addr[2] == ip[2] &&
			arp_cache[i].ip_addr[1] == ip[1] &&
			arp_cache[i].ip_addr[0] == ip[0] )
			return arp_cache[i].hw_addr;

	/* not on cache, do a request */
	do_arp_request(ip);

	return NULL;
}

int send_packet(uint8_t *packet, uint16_t length)
{
	struct ip_packet *frame = (struct ip_packet *) packet;
	uint8_t * target_hw_addr;
	
	if (((frame->target_ip_addr[0] & netmask[0]) != subnet[0]) ||
			((frame->target_ip_addr[1] & netmask[1]) != subnet[1]) ||
			((frame->target_ip_addr[2] & netmask[2]) != subnet[2]))
	{
		/* out of subnet, send it to gateway */
		target_hw_addr = resolve_ip(gateway);
	}
	else
	{
		/* we are on same subnet, resolve target IP address */
		target_hw_addr = resolve_ip(frame->target_ip_addr);
	}

	/* address not yet on cache, come back later */
	if (target_hw_addr == NULL)
	{
		memcpy(tx_buffer.data, packet, length);
		tx_buffer.length = length;
		tx_buffer.valid = 1;
		return 1;
	}

	memcpy(ethernet_header, target_hw_addr, 6);
	writeFrame(packet, length);

	return 0;
}

#ifndef UZIX_MODULE
	
void do_icmp_reply (uint8_t *data, uint16_t size)
{
	struct ip_packet *packet = (struct ip_packet *)data;
	/*add_arp_entry(data + 6, data + 0x1a);*/

	memcpy(packet->target_ip_addr, packet->source_ip_addr, 4);
	memcpy(packet->source_ip_addr, ip_addr, 4);
	data[20] = 0; /* PING reply */

	send_packet(data, size);
}

#endif


uint8_t * getPacket(uint16_t * length)
{
#asm
	; poll for a received packet
	setRegister 0124h

	in a, (0Dh)
	and 1
        jp z, fimzis

	; reread RxStatus
	in a, (1)
	in a, (0)
#endasm

	/* this reads the frame */
	/* fix it: make it all ASM */
	rx_buffer.length = getFrame(rx_buffer.data);
	*length = rx_buffer.length - 14;

	switch ( rx_buffer.data[13] ) /* FIXME: check the whole 16-bit number */
	{
		case 0x00:
			return rx_buffer.data + 14;

		case 0x06: /* ARP request */
			switch (rx_buffer.data[0x15])
			{
				case 1:		/* ARP request */
					do_arp_reply(&rx_buffer);
					break;

				case 2:		/* ARP reply */
					add_arp_entry(rx_buffer.data + 0x16, rx_buffer.data + 0x1c);
					break;

				default:
					break;
			}

			break;

		case 0x35:	/* Invalid packet */
#ifndef UZIX_MODULE
			printf("RARP\n");
#endif
			break;

		default:	/* Invalid packet */
			break;
	}
#asm
fimzis:
#endasm
	return NULL;
}

void initialize_eth(unsigned char *ip, unsigned char *netm, unsigned char *gate)
{
	memcpy(ip_addr, ip, 4);
	memcpy(netmask, netm, 4);
	memcpy(gateway, gate, 4);
	
	subnet[0] = ip_addr[0] & netmask[0];
	subnet[1] = ip_addr[1] & netmask[1];
	subnet[2] = ip_addr[2] & netmask[2];
	subnet[3] = ip_addr[3] & netmask[3];

	init_cs8900();

	/* resolve the gateway right away */
	do_arp_request(gateway);
	
	/* FIXME: select correct size here */
	rx_buffer.data = (uint8_t *) malloc(1518);
	tx_buffer.data = (uint8_t *) malloc(1518);
}

#ifndef UZIX_MODULE

int main()
{
	uint16_t foo;
	uint8_t *pa;

	printf ("CS8900A driver for Uzix v0.1\n");
	printf ("(c)2003 Banana Enterprises\n\n");

	if (detect_cs8900())
	{
		fprintf(stderr, "Error: no CS8900A found\n");
		return 1;
	}

	printf ("Using defaults: IP address\t192.168.1.9\nNetmask\t255.255.255.0\nGateway\t192.168.1.1\n");
	initialize_eth((uint8_t *)"\xc0\xa8\x01\x09", (uint8_t *)"\xff\xff\xff\xff", (uint8_t *)"\xc0\xa8\x01\x01");

	for (;;) {
		_ETH_mainloop();
		pa = _ETH_getpacket(&foo);
		if (pa) {
			if (pa[0x17-14] == 1) {
				do_icmp_reply(pa, foo);
				free(pa);
			}
		}
	}
}

#endif

void _ETH_mainloop() {
	uint16_t size;
	uint8_t *pack, *pack2;
	
	if (!(pack = getPacket(&size))) return;
	/* buffer full; discard */
	if (((rec_buff_last + 1) & (MAX_REC_BUF-1)) == rec_buff_idx) return;
	/* no memory; discard */
	if (!(pack2 = malloc(size))) return;
	memcpy(pack2, pack, size);
	rec_buff[rec_buff_last] = pack2;
	rec_buff_sizes[rec_buff_last] = size;
	rec_buff_last = (++rec_buff_last & (MAX_REC_BUF-1));
}

uint8_t *_ETH_getpacket(uint16_t *size) {
	uint8_t *pack;

	if (rec_buff_last == rec_buff_idx) return NULL;
	*size = rec_buff_sizes[rec_buff_idx];
	pack = rec_buff[rec_buff_idx];
	rec_buff_idx = (++rec_buff_idx & (MAX_REC_BUF-1));
	/* caller is responsible for freeing the buffer with free()! */
	return pack;
}

#asm
_detect_cs8900:
	ld a, 0		; set packet page pointer to 0x0000
	out (PPPntrH),a
	out (PPPntrL),a
	
	readData

	; compare value
	ld hl, 0E63h
	sbc hl, de

	ret

_init_cs8900:
; Set the MAC address
	setRegister 0158h	; Internal Address Register
	writeData 0DEF0h
	
	setRegister 015Ah	; Internal Address Register
	writeData 000BAh
	
	setRegister 015Ch	; Internal Address Register
	writeData 0100h

; allow packets
	setRegister 0104h

	ld a, 00001101B
	out (0dh), a

; enable receiver and transmitter
	setRegister 0112h

	ld a, 11000000B		; set SerTxOn and SerRxOn
	out (0ch), a

	ret

_getFrame:
        getParInHL      ; pointer to buffer  

; get frame length
	in a, (1)
	ld d, a
	in a, (0)
	ld e, a

; save length and divide it by two
	ld b, d
	ld c, e
	srl b
	rr c

; load the frame into memory
1:	in a, (1)
	ex af,af'
	in a, (0)

	ld (hl), a
	inc hl
	ex af,af'
	ld (hl), a
	inc hl

	dec bc
	ld a, b
	or c

	jr nz, 1b

	ld h, d
	ld l, e
	ret

; sendframe
;
; arguments:
;	DE	pointer to packet_data
;	BC	size of packet
;
; returns:
;	nothing

_sendFrame:
        getParInDE_BC   ; buffer location, TxLength

; bid for buffer space by writing command and frame length
	ex de,hl
	ld c, 04h
	ld a, 0c0h	; write 00C0h command to TxCMD
	out (c), a
	inc c
	ld a, 0
	out (c), a
	inc c		; write TxLength
	out (c), e
	inc c
	out (c), d
	srl d
	rr e
; poll for space
	setRegister 0138h	; BusST register
1:	in a, (0dh)
	and 1
	jp z, 1b

; write loop
2:	ld c, 0
	outi
	ld c, 1
	outi

	dec de
	ld a,d
	or e
	jr nz, 2b
	ret

; writeframe
;
; this function writes the ethernet_header and the packet
;

_writeFrame:
; bid for buffer space by writing command and frame length
        getParInDE_BC   ; buffer location, TxLength

	push hl

	ld bc, 14
	add hl, bc	; add header length
	
	ld c, 04h
	ld a, 0c0h	; write 00C0h command to TxCMD
	out (c), a
	inc c
	ld a, 0
	out (c), a

	inc c		; write TxLength
	out (c), l
	inc c
	out (c), h

; poll for space
	setRegister 0138h	; BusST register

1:	in a, (0dh)
	and 1
	jp z, 1b

; write header
	ld c, 7
	ld hl, _ethernet_header

2:	ld a, (hl)
	out (0), a
	inc hl

	ld a, (hl)
	out (1), a
	inc hl

	dec c
	jr nz, 2b

	pop hl

; write loop
	srl h
	rr l

3:	ld a, (de)
	out (0), a
	inc de

	ld a, (de)
	out (1), a
	inc de

	dec hl

	ld a,h
	or l

	jr nz, 3b
	ret

#endasm
