typedef unsigned int uint16_t;
typedef unsigned char uint8_t;

#define MAC_ADDRESS_SIZE 6
#define IP_ADDRESS_SIZE 4
#define ARP_CACHE_SIZE 5

struct ethernet_frame {
	uint16_t length;
	uint8_t *data;
	uint8_t valid;	/* used only for Tx buffer */
};

struct arp_packet {
	uint16_t hw_type;
	uint16_t protocol;
	uint8_t hw_size;
	uint8_t protocol_size;
	uint16_t opcode;
	uint8_t source_hw_addr[MAC_ADDRESS_SIZE];
	uint8_t source_ip_addr[IP_ADDRESS_SIZE];
	uint8_t target_hw_addr[MAC_ADDRESS_SIZE];
	uint8_t target_ip_addr[IP_ADDRESS_SIZE];
};

struct ip_packet {
	uint8_t version;
	uint8_t ToS;
	uint16_t length;
	uint16_t identification;
	uint16_t fragment_offset;
	uint8_t TTL;
	uint8_t protocol;
	uint16_t checksum;
	uint8_t source_ip_addr[IP_ADDRESS_SIZE];
	uint8_t target_ip_addr[IP_ADDRESS_SIZE];
	uint8_t *data;
}
struct arp_cache_entry {
	uint8_t hw_addr[MAC_ADDRESS_SIZE];
	uint8_t ip_addr[IP_ADDRESS_SIZE];
	uint8_t timer;
};

uint8_t free_entry = 0;

struct arp_cache_entry arp_cache[ARP_CACHE_SIZE];

uint16_t detect_cs8900();
void init_cs8900();
uint16_t getRxLength();
uint16_t getFrame(uint8_t *);
void sendFrame(uint8_t *, uint16_t);
void initialize_eth(unsigned char *ip, unsigned char *netm, unsigned char *gate);
uint8_t *_ETH_getpacket(uint16_t *size);
void _ETH_mainloop();
uint8_t mytime = 0;

#ifdef FUDEBATOR
#asm
MACRO   getParInHL
        ex de,hl
ENDM

MACRO   getParInDE_BC
        ; yep! nothing!
ENDM
#endasm
#else
#asm
MACRO   getParInHL
	pop de
	pop hl		; pointer to buffer
	push hl
	push de
ENDM

MACRO   getParInDE_BC
	pop bc
	pop de		; buffer location
	pop hl		; TxLength
	push hl
	push de
	push bc
ENDM

#endasm
#endif

#asm
PPPntrH		EQU	0Bh
PPPntrL		EQU	0Ah

MACRO	setRegister, addr
	ld a, .high. addr
	out (0Bh), a
	ld a, .low. addr
	out (0Ah), a
ENDM

MACRO	writeData, data
	ld a, .high. data
	out (0Dh), a
	ld a, .low. data
	out (0Ch), a
ENDM

MACRO	readData
	in a,(0Dh)	; loads data
	ld e, a
	in a,(0Ch)
	ld d, a
ENDM
#endasm
